#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <vga.h>
#include <scanner.h>

#define MIN(a, b)	((a)<(b)?(a):(b))

#define MAXMATRIX 8

void usage(void)
{ printf("Usage : scandemo [-gx]\n");
  printf(" -gx set greyscale-matrix size (1-8)\n");
}

int main(int argc,char *argv[])
{
  int n, n1, fd, r, sl, ht, wd;
  int greymatrix=1;
  struct modeinfo_struct mi;
  struct scanner_type sct;
  unsigned char line[1280], mask;
  unsigned char buf[MAXMATRIX*210];

  for(n=1;n<argc;n++)
  { if (*argv[n]!='-') usage();
    switch(argv[n][1])
    {case 'g':greymatrix=argv[n][2]-'0';
    	      if (greymatrix>8||greymatrix<1) usage();
    	      break;
     default :usage();break;
    }
  } 
  if ((fd = open("/dev/scan", O_RDONLY)) < 0) {
    perror("open");
    exit(1);
  }
  if ((ioctl(fd, HSIOCGSCT, &sct)) < 0) {
    perror("ioctl");
    close(fd);
    exit(2);
  }
  printf("Scanner type: Manufacturer : '%s'\n", sct.manufacturer);
  printf("              Model        : '%s'\n", sct.model);
  if ((ioctl(fd, HSIOCGMOD, &mi)) < 0) {
    perror("ioctl");
    close(fd);
    exit(2);
  }
  printf("Scanner set to %d dpi (%d bytes per scanline)\n", mi.dpi, mi.bpl);

  vga_init();
  if ((r = vga_getdefaultmode()) < 0) {
    fprintf(stderr, "GSVGAMODE environment variable contains no valid mode." \
	    "Please change this.\n");
    close(fd);
    exit(3);
  }
  vga_setmode(r);
  if (vga_getcolors() != 256) {
    fprintf(stderr, "Please select a 256 color mode\n");
    vga_setmode(TEXT);
    close(fd);
    exit(4);
  }
  wd = vga_getxdim();
  ht = vga_getydim();

  for (n = 0; n < 256 && n <= greymatrix*greymatrix ;n++)
  { n1=n*63/greymatrix/greymatrix;
    vga_setpalette(n, n1, n1, n1); }
  
  while (1)
    for (sl = 0; sl < ht;) {
      for (n = 0; n < sizeof(line); n++) line[n] = 0;
      if ((r = read(fd, &buf, mi.bpl*greymatrix)) < 0) {
	vga_setmode(TEXT);
	perror("read");
	close(fd);
	exit(1);
      } else if (r > 0) {
	for (n=0;n<mi.bpl*greymatrix;n++) 
	{ if ((n%mi.bpl)==0) n1=-1;
	  for(r=0;r<8;r++) 
	  { if (((n*8+r)%greymatrix)==0) ++n1;
	    if (buf[n]&0x80) line[n1]++;
	    buf[n]<<=1;
	  }
	}
	vga_drawscanline(sl++, line);
      }
    }
}
