/*
 *  logi32.c - Scanner-driver for Logitech SCANMAN+
 *
 *  Copyright (c) 1994 Andreas Beck (becka@hp.rz.uni-duesseldorf.de)
 *  Parts copyright (c) 1994 Thomas Faehnle (Thomas.Faehnle@student.uni-ulm.de)
 *
 *  This is the driver for the LOGITECH SCANMAN+ 400 dpi halftone handheld 
 *  scanner driven thru a LOGITECH interface . IRQ and DMA channel are
 *  adjusted thru software, they are defined in scanner.h.
 */

/* 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  logi32.c,v 0.0.1 1994/04/26 20:58:42
 */
 
#include "logi32.h"
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>	/* for f_flags bits */
#include <linux/module.h>
#define REALLY_SLOW_IO
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/segment.h>

#define MODEINFO	logi_modeinfo[modeindex]
#define SIGNAL_OCCURED	(current->signal & ~current->blocked)
#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define RD_BLKEND	(rd.blkptr->blkoff + rd.blkptr->blksiz)
#define WR_BLKEND	(wr.blkptr->blkoff + wr.blkptr->blksiz)
#define RD_BUFSPC	(wr.index >= rd.index ? ((wr.index == rd.index) && !ds.buffer_empty ? \
	bufsiz : 0) + wr.index - rd.index : bufsiz - rd.index + wr.index)
#define WR_BUFSPC	(wr.index < rd.index ? rd.index - wr.index : \
	((wr.index == rd.index) && !ds.buffer_empty ? 0 : bufsiz) - wr.index + rd.index)
#define RD_ADDR		(&(rd.blkptr->data[rd.index - rd.blkptr->blkoff]))
#define WR_ADDR		(&(wr.blkptr->data[wr.index - wr.blkptr->blkoff]))
#define SELECT_TH	(MODEINFO.bpl * select_threshold)

#ifdef DEBUG_LOGI32
#define PRINTK(x...)	{if(debug) printk(x);}
#else
#define PRINTK(x...)
#endif

static struct { int type,ready_mask,tf2;} sct={2,0x40,0};
static struct scanner_type sctxt={"LOGITECH","SCANMAN+"};
static struct wait_queue *wq = NULL;
static int modeindex = 0;
static struct device_status_struct ds = { 0, 0, 0, 0 };
static struct buffer_block_struct *memroot = NULL;
static int buflines = 0, req_lines = 0, bufsiz = 0, bytes_missing = 0;
static struct buffer_pointer rd = { NULL, 0 }, wr = { NULL, 0 };
static int select_threshold = 1;
#ifdef DEBUG_LOGI32
static int debug;
#endif

char kernel_version[]=KERNEL_VERSION;	/* needed for module installer */

/************************************************************************/

static void wait_ready(void)		/* Wait until Scanner is ready */
{ while(inb(LOGI32_IOBASE)&sct.ready_mask);
  if (sct.type==1)
  { while(inb(LOGI32_IOBASE)&0x40) inb(LOGI32_IOBASE+2); }
  else
  { while(inb(LOGI32_IOBASE)&0x80) 
    { outb(1,LOGI32_IOBASE);inb(LOGI32_IOBASE+1); }
  }
}

static void light_on(void)		/* Turn on the LEDs */
{ outb(0,LOGI32_IOBASE);
  outb(inb(LOGI32_IOBASE+1)|0x01,LOGI32_IOBASE+1);
  /* once upon a time here was a delay ... is it needed ? */
}

static void light_off(void)		/* Turn them off */
{ outb(0,LOGI32_IOBASE);
  outb(inb(LOGI32_IOBASE+1)&0xfe,LOGI32_IOBASE+1);
}

static int get_bytes_per_line(void)	/* see left :-) */
{ static int data_39=8;
  if (sct.type!=1) 
  { if (sct.tf2)
    { outb(0,LOGI32_IOBASE);
      outb(inb(LOGI32_IOBASE+1)|0x02,LOGI32_IOBASE+1);
      outb(8,LOGI32_IOBASE);
      outb(2,LOGI32_IOBASE+1);}

    light_on();
    if (sct.type==3)
    { outb(10,LOGI32_IOBASE);
      data_39=(((inb(LOGI32_IOBASE+1)^0xff)&0x60)>>3)+4;
    } 
    else
    { light_off();
      outb(0,LOGI32_IOBASE);
      data_39=(inb(LOGI32_IOBASE+1)&0x0c)+4;
    }
  }
  /* Here comes what was SUB_get_bps in DOS */
  if (sct.type==3)
  { /* FIXME : This should be filled ... t3bps ...*/ }

  /* Here comes what was SUB_calc_by_per_scanl in DOS */
  /* FIXME : This should be prepared to handle other sizes and t3-modes ! */

  return((0x67*data_39)/8);
}

static void start_scanning(void)	/* <== */
{ if (sct.type==1)
  { outb(6,LOGI32_IOBASE);}
  else
  { light_on();
    if (sct.tf2)
    { outb(1,LOGI32_IOBASE);inb(LOGI32_IOBASE+1);
      outb(1,LOGI32_IOBASE);inb(LOGI32_IOBASE+1);
    }
  }
  if (sct.type==3)
  { outb(0x11,LOGI32_IOBASE);inb(LOGI32_IOBASE+1); 
    /* FIXME: this sets a flag depending on inb-result ...*/}

  if (sct.type==1)
  { outb(MODEINFO.bpl,LOGI32_IOBASE+6); /* Setup transfer width */
  } else
  { outb(8,LOGI32_IOBASE);outb( MODEINFO.bpl    &0xff,LOGI32_IOBASE+1);
    outb(9,LOGI32_IOBASE);outb((MODEINFO.bpl>>8)&0xff,LOGI32_IOBASE+1);
  }
  wait_ready();

  if (sct.type==1)
  { inb(LOGI32_IOBASE+3); } 
  else
  { outb(5,LOGI32_IOBASE);inb(LOGI32_IOBASE+1); }
}

static int logi32_get_mode(void)
{ int bpl;
  logi_modeinfo[0].bpl=bpl=get_bytes_per_line();
  for(modeindex=(sizeof(logi_modeinfo)/sizeof(logi_modeinfo[0]))-1;
      modeindex>0;modeindex--)
   if (logi_modeinfo[modeindex].bpl==bpl) break;
  return(0);
}

static __inline__ int dma_terminal_count(unsigned int channel)
{
  if(channel <= 3)
    return inb(DMA1_STAT_REG) & (1 << channel);
  else
    return inb(DMA2_STAT_REG) & (1 << (channel & 3));
}

static __inline__ void setup_xfer(char *addr)
{
  disable_dma(LOGI32_DMA);
  set_dma_mode(LOGI32_DMA, DMA_MODE_READ);
  clear_dma_ff(LOGI32_DMA);
  set_dma_addr(LOGI32_DMA, (int) addr);
  set_dma_count(LOGI32_DMA, MODEINFO.bpl);
  enable_dma(LOGI32_DMA);
}

static void scanner_standby(void)
{
  ds.irq_enabled=0;
  if (sct.type!=1) /* deactivate hardware */
  { outb(0,LOGI32_IOBASE);
    outb(0xfd&inb(LOGI32_IOBASE+1),LOGI32_IOBASE+1);
  }  				/* gives me a safer feeling if the */
  disable_dma(LOGI32_DMA);	/* DMA channel is also disabled on the 8237 */

  if (sct.type==1) /* deactivate hardware */
  { outb(2,LOGI32_IOBASE);}
  else
  { light_off();
    outb(3,LOGI32_IOBASE);
    outb(0x77&inb(LOGI32_IOBASE+1),LOGI32_IOBASE+1);
    outb(4,LOGI32_IOBASE);outb(0x10,LOGI32_IOBASE+1);
  }
}

static void scanner_operate(void)
{
  start_scanning();
  if (sct.type!=1)
  { outb(4,LOGI32_IOBASE);outb(0,LOGI32_IOBASE+1);
    outb(3,LOGI32_IOBASE);
    outb(0x88|inb(LOGI32_IOBASE+1),LOGI32_IOBASE+1);/* 0x8 is IRQ-enable */
    light_on();
  }

  cli();
  setup_xfer(WR_ADDR);
  ds.irq_enabled = 1;
  sti();

  enable_dma(LOGI32_DMA);
  if (sct.type==1) /* activate the hardware */
  { outb(0x1e,LOGI32_IOBASE); /* 0x10 is IRQ-Flag */
  } else
  { outb(0,LOGI32_IOBASE);
    outb(0x02|inb(LOGI32_IOBASE+1),LOGI32_IOBASE+1);
  }
}
      
static void logi32_free_mem(void)
{
  struct buffer_block_struct *tmp = memroot, *next;
  unsigned int size;
  if(!buflines) {
    PRINTK("logi32_free_mem: there's nothing to free\n");
    return;
  }
  scanner_standby();	/* make sure DMA xfers are really off */
  while(tmp) {
    size = tmp->blksiz + HEADERSIZ;
    next = tmp->next;
    kfree_s((char*) tmp, size);
    PRINTK("kfree_s'd %d bytes (data+header) at 0x%08x\n", size, (int) tmp);
    tmp = next;
  };
  memroot = NULL; rd.blkptr = NULL; wr.blkptr = NULL;
  buflines = 0; bufsiz = 0;
}

static int logi32_get_mem(int lines)
{
  struct buffer_block_struct *tmp, **lastptr;
  unsigned int blksiz;
  if((lines * MODEINFO.bpl) > MAX_BUFSIZ)
    return -EINVAL;
  if(buflines)
    logi32_free_mem();
  lastptr = &memroot;
  while(lines > 0) {
    blksiz = (((lines * MODEINFO.bpl) > MAX_BLK) ? \
	(MAX_BLK - MAX_BLK % MODEINFO.bpl) : (lines * MODEINFO.bpl));
    tmp=(struct buffer_block_struct*)kmalloc(blksiz + HEADERSIZ, GFP_KERNEL);
    if(!tmp) {
      logi32_free_mem();
      return -ENOMEM;
    }
    tmp->next = NULL;
    *lastptr = tmp;
    lastptr = &(tmp->next);
    tmp->blksiz = blksiz;
    tmp->blkoff = buflines * MODEINFO.bpl;
    lines -= blksiz / MODEINFO.bpl;
    buflines += blksiz / MODEINFO.bpl;
    bufsiz += blksiz;
    PRINTK("kmalloc'ed %d bytes (data+header) at 0x%08x, %d lines, remaining: %d, offset: %d\n", \
    	blksiz+HEADERSIZ, (int) tmp, buflines, lines, tmp->blkoff);
    rd.blkptr = memroot; rd.index = 0;
    wr.blkptr = memroot; wr.index = 0;
    ds.buffer_empty = 1;
  }
  return 0;
}
     
static __inline__ int get_available_bytes(char **user_buf, int requested)
{
  int bytes_to_copy;
  int chunksiz, bytes_copied = 0;
  cli(); bytes_to_copy = MIN(requested, RD_BUFSPC); sti();
  PRINTK("get_available_bytes: %d bytes to copy (%d requested, RD_BUFSPC == %d)\n", \
  	bytes_to_copy, requested, RD_BUFSPC);
  PRINTK("get_available_bytes: rd.index == %d, wr.index == %d\n", rd.index, wr.index);
  while(bytes_to_copy > 0) {
    cli(); chunksiz = MIN(bytes_to_copy, RD_BLKEND - rd.index); sti();
    PRINTK("get_available_bytes: copying chunk of %d bytes to user space\n", chunksiz);
    PRINTK("get_available_bytes: RD_BLKEND is %d\n", RD_BLKEND);
    memcpy_tofs(*user_buf, RD_ADDR, (long) chunksiz);
    *user_buf += chunksiz;
    rd.index += chunksiz;
    bytes_copied += chunksiz;
    bytes_to_copy -= chunksiz;
    if(rd.index >= RD_BLKEND) {
      rd.blkptr = rd.blkptr->next;
      if(!rd.blkptr) {	/* last block in linked list */
	rd.blkptr=memroot;
	rd.index -= bufsiz;
      }
    }
  }
  cli(); if(rd.index == wr.index) ds.buffer_empty = 1; sti();
  PRINTK("get_available_bytes: copied total of %d bytes to user space\n", bytes_copied);
  return bytes_copied;
}

static int logi32_read(struct inode *inode, struct file *file, \
	char *buf, int requested)
{
  int r, tmp = requested;
  if(!buflines && ((r=logi32_get_mem(req_lines = DEFAULT_BUFLINES)) < 0))
    return r;
  requested -= get_available_bytes(&buf, requested);
  if(!ds.irq_enabled && (WR_BUFSPC >= MODEINFO.bpl)) {
    PRINTK("logi32_read: I see some space in my buffer, will enable irq\n");
    scanner_operate();
  }
  if(file->f_flags & O_NONBLOCK) {
    PRINTK("logi32_read: reading in O_NONBLOCK mode\n");
    return tmp - requested;
  }
  while(requested > 0) {
    bytes_missing = requested;
    PRINTK("logi32_read: %d bytes are still missing, I'll sleep a bit. ds.irq_enabled == %d\n", \
    	bytes_missing, ds.irq_enabled);
    if(!ds.irq_enabled) {
      PRINTK("logi32_read: re-enabling interrupts\n");
      scanner_operate();
    }
    ds.process_sleeps = 1;
    interruptible_sleep_on(&wq);
    requested -= get_available_bytes(&buf, requested);
    if(SIGNAL_OCCURED)
      return tmp - requested;
  }
  return tmp;
}

static int logi32_select(struct inode *inode, struct file *file, \
	int type, select_table *st)
{
  int tmp, r;
  if(type != SEL_IN)
    return 0;
  if(!buflines && ((r=logi32_get_mem(req_lines = DEFAULT_BUFLINES)) < 0))
    return r;
  cli();
  tmp=RD_BUFSPC;
  sti();
  if(tmp >= SELECT_TH) {
    PRINTK("logi32_select: device is ready (%d bytes available)\n", tmp);
    return 1;			/* device is ready */
  }
  PRINTK("logi32_select: device NOT ready\n");
  if(st) {
    PRINTK("logi32_select: waiting for device to get ready\n");
    bytes_missing = SELECT_TH;
    ds.process_sleeps = 1;
    if(!ds.irq_enabled) {
      PRINTK("logi32_select: IRQ not enabled -- will turn on scanner\n");
      scanner_operate();
    }
  }
  select_wait(&wq, st);
  return 0;
}

static void logi32_interrupt(int unused_pointer_to_pt_regs_struct)
{
  static int bytes_til_wakeup;
  int help;
  if(!ds.irq_enabled) {
    PRINTK("logi32.c: logi32_interrupt: unexpected interrupt\n");
    return;
  }
  if(sct.type==1)
  { if (((help=inb(LOGI32_IOBASE))&0x20)==0) 
    { PRINTK("logi32.c: Interrupt didn't originate from scanner\n");return;}
    outb(help&0xdf,LOGI32_IOBASE);	/* Clear int mask ... */
  } else
  { outb(4,LOGI32_IOBASE);
    if (((help=inb(LOGI32_IOBASE+1))&0x01)==0) 
    { PRINTK("logi32.c: Interrupt didn't originate from scanner\n");return;}
    if (help&0x02) 
    { outb((help|0x10)&0xfd,LOGI32_IOBASE+1);}
    else 
    { if (help&0x10) PRINTK("Flag activated !\n");
      outb(0,LOGI32_IOBASE+1);PRINTK("Returning ... why ?\n");
      return;
    }
  }
  if(!dma_terminal_count(LOGI32_DMA)) {
    PRINTK("logi32.c: logi32_interrupt: DMA transfer not finished yet\n");
    return;
  }
  PRINTK("logi32_interrupt: I have been called, bytes_missing == %d\n", bytes_missing);

  if(bytes_missing > 0) {
    bytes_til_wakeup = bytes_missing;
    bytes_missing = 0;
  }
  wr.index += MODEINFO.bpl;
  if(wr.index >= WR_BLKEND) {
    wr.blkptr = wr.blkptr->next;
    if(!wr.blkptr) {
      wr.blkptr = memroot;
      wr.index -= bufsiz;
    }
  }
  if(wr.index == rd.index)
    ds.buffer_empty = 0;
  PRINTK("logi32_interrupt: wr.index is now %d\n", wr.index);
  if(bytes_til_wakeup > 0)
    bytes_til_wakeup -= MODEINFO.bpl;
  if(bytes_til_wakeup <= 0)
    if(ds.process_sleeps) {
      PRINTK("logi32_interrupt: the requested bytes are there, will wake up process\n");
      wake_up_interruptible(&wq);
      ds.process_sleeps = 0;
    }
  if(WR_BUFSPC >= MODEINFO.bpl) {
    PRINTK("logi_interrupt: there's still free buffer space (%d bytes), irq stays enabled\n", \
    	WR_BUFSPC);
  } else {
    PRINTK("logi32_interrupt: buffer is full -- turning off irq\n");
    scanner_standby();
    if(ds.process_sleeps) {
      wake_up_interruptible(&wq);
      ds.process_sleeps = 0;
    }
    return;
  }

  if (sct.type==2)
  { while (inb(LOGI32_IOBASE)&0x80)
    {outb(1,LOGI32_IOBASE);inb(LOGI32_IOBASE+1);}
  }
  if (sct.type!=1)
  { outb(3,LOGI32_IOBASE);inb(LOGI32_IOBASE); }

  if (sct.type==1)
  { while(inb(LOGI32_IOBASE)&0x01);
    outb(6,LOGI32_IOBASE);}

  if (sct.type!=1)
  { outb(3,LOGI32_IOBASE);
    outb(inb(LOGI32_IOBASE+1)&0x7f,LOGI32_IOBASE+1);
  }
  setup_xfer(WR_ADDR);
  enable_dma(LOGI32_DMA);
  if (sct.type==1) /* activate the hardware */
  { outb(0x1e,LOGI32_IOBASE); /* 0x10 is IRQ-Flag */
  } else
  { outb(4,LOGI32_IOBASE);outb(0x10,LOGI32_IOBASE+1);
    outb(3,LOGI32_IOBASE);
    outb(0x80|inb(LOGI32_IOBASE+1),LOGI32_IOBASE+1);
    outb(0,LOGI32_IOBASE);
    outb(0x02|inb(LOGI32_IOBASE+1),LOGI32_IOBASE+1);
  }
  return;
}

static int logi32_ioctl(struct inode *inode, struct file *file, \
	unsigned int cmd, unsigned long arg)
{
  int r, s;
  struct registers_struct rs;
  if(!arg)
    return -EINVAL;
  switch(cmd) {
    case HSIOCGSCT:			/* get scanner_type struct */
      memcpy_tofs((char*) arg, (char*) &sctxt, (long) sizeof(struct scanner_type));
      return 0;
    case HSIOCGMOD:			/* get modeinfo struct */
      if((r=logi32_get_mode()) >= 0)
	memcpy_tofs((char*) arg, (char*) &MODEINFO, (long) sizeof(struct modeinfo_struct));
      return r;
    case HSIOCSMOD:			/* set modeinfo struct */
      modeindex=0;			/* Do not overwrite a const ... */
      memcpy_fromfs((char*) &MODEINFO, (char*) arg, (long) sizeof(struct modeinfo_struct));
      return 0;
    case HSIOCGREG:			/* get setting of misc registers */
      memcpy_fromfs((char*) &rs, (char*) arg, (long) sizeof(struct registers_struct));
      if (rs.regnum<0||rs.regnum>0x0f) return -EACCES;
      rs.data=inb_p(LOGI32_IOBASE+rs.regnum);
      memcpy_tofs((char*) arg, (char*) &rs, (long) sizeof(struct registers_struct));
      return 0;
    case HSIOCSREG:			/* set misc registers */
      memcpy_fromfs((char*) &rs, (char*) arg, (long) sizeof(struct registers_struct));
      if (rs.regnum<0||rs.regnum>0x0f) return -EACCES;
      outb_p(rs.data, LOGI32_IOBASE+rs.regnum);
      return 0;
    case HSIOCSBUF:			/* set size of internal buffer */
      if(!(r=get_fs_long(arg)))
	logi32_free_mem();
      else {
	if((s=logi32_get_mode()) < 0)
	  return s;
	else {
	  req_lines = r;
	  logi32_get_mem(req_lines);
	}
      }
      return 0;
    case HSIOCGBUF:			/* get current buffer size */
      put_fs_long((unsigned long) req_lines, arg);
      return 0;
    case HSIOCSSTH:			/* set select(2) threshold */
      select_threshold=get_fs_long(arg);
      return 0;
    case HSIOCGSTH:			/* get current select(2) threshold */
      put_fs_long((unsigned long) select_threshold, arg);
      return 0;
    default:
      return -EINVAL;
  }
}

static int logi32_open(struct inode *inode, struct file *file)
{
  int r;

#ifndef DEBUG_LOGI32
  if(MINOR(inode->i_rdev) != 0)
    return -ENODEV;
#else
  debug=MINOR(inode->i_rdev);
#endif

  if(ds.device_open)
    return -EBUSY;

  ds.irq_enabled = 0; ds.process_sleeps = 0;
  logi32_get_mode();

  if((r=request_irq(LOGI32_IRQ, logi32_interrupt)))	/* request IRQ */
    return r;

  if((r=request_dma(LOGI32_DMA))) {
    free_irq(LOGI32_IRQ);
    return r;
  }

  scanner_standby();
  MOD_INC_USE_COUNT;
  ds.device_open = 1;
  return 0;	/* device open, scanner is in standby mode now */
}

static void logi32_release(struct inode *inode, struct file *file)
{
  scanner_standby();
  wait_ready();
  if(buflines)
    logi32_free_mem();
  free_dma(LOGI32_DMA);
  free_irq(LOGI32_IRQ);
  ds.device_open = 0;
  MOD_DEC_USE_COUNT;
}

static int logi32_lseek(struct inode *inode, struct file *file, \
	off_t offset, int origin)
{
  return -ESPIPE;	/* in case some luser tries to seek on a scanner... */
}

static struct file_operations logi32_fops = {
  logi32_lseek,
  logi32_read,
  NULL,		/* write */
  NULL,		/* readdir */
  logi32_select,
  logi32_ioctl,
  NULL,		/* mmap */
  logi32_open,
  logi32_release,
};

static int check_4_scanner(void)
{ int help,tf,t3f;
  static char *scname[4]={"UNKNOWN","SCANMAN","SCANMAN+","NCRSCAN"};
  tf=0;
  t3f=1;
  help=inb(LOGI32_IOBASE+7)&0x0f;
  if (help==1)
  { sct.ready_mask=0x01;
    sct.type=1;
  } else
  { sct.ready_mask=0x40;
    outb(5,LOGI32_IOBASE);		/* Get reg #5 */
    help=inb(LOGI32_IOBASE+1)&0x0f;
    switch(help)
    {case 0x0:
     case 0xf:printk("Scanner-interface not detected.\n");return(-ENODEV);
     case 0x2:sct.type=2;break;
     case 0x8:sct.type=2;tf=sct.tf2=1;break;
     default :printk("Scanner-interface is incompatible.\n");return(-ENODEV);
    }
  }
  if (sct.type!=1)
  { outb(0,LOGI32_IOBASE);
    outb(inb(LOGI32_IOBASE+1)&0x7f,LOGI32_IOBASE+1);
    outb(0,LOGI32_IOBASE);
    outb(inb(LOGI32_IOBASE+1)|0x20,LOGI32_IOBASE+1);
    outb(0,LOGI32_IOBASE);
    if (inb(LOGI32_IOBASE+1)&0x20) tf=1;
    light_on();
    outb(5,LOGI32_IOBASE);
    if (inb(LOGI32_IOBASE+1)&0x10) t3f=0;
    light_off();
  }
  if (tf==1&&t3f==0) sct.type=3;
  if (sct.type!=1)
  { outb(2,LOGI32_IOBASE);
    outb(0x42,LOGI32_IOBASE+1);
    outb(3,LOGI32_IOBASE);
    switch(LOGI32_IRQ)
    {case  2:
     case  9:help=0x04;break;
     case  3:help=0x01;break;
     case  4:help=0x02;break;
     case  5:help=0x03;break;
     case  7:help=0x05;break;
     case 11:help=0x06;break;
     case 12:help=0x07;break;
    }
    outb((LOGI32_DMA==1 ? 0x10 : 0x30)|help,LOGI32_IOBASE+1);
    outb(0,LOGI32_IOBASE);
    outb(inb(LOGI32_IOBASE+1)&0xfd,LOGI32_IOBASE+1);
    if (sct.tf2==0) {outb(6,LOGI32_IOBASE);outb(0x0e,LOGI32_IOBASE+1);}
    outb(4,LOGI32_IOBASE);outb(0,LOGI32_IOBASE+1);
    outb(0xa,LOGI32_IOBASE);outb(0,LOGI32_IOBASE+1);
  }
  if (sct.type<0||sct.type>3) sct.type=0; /* Unknown scanner,continue ...*/
  help=0;
  while(help<sizeof(sctxt.model) &&
        (sctxt.model[help]=scname[sct.type][help])) help++;
  printk("check_4_scanner detected %s scanner.\n",sctxt.model);
  return(0);/* Scanner is there and operative */
}

int init_module(void)
{
  printk("LOGITECH handheld scanner driver  version 0.1\n");
  if(register_chrdev(LOGI32_SCANNER_MAJOR, LOGI32_SCANNER_NAME, &logi32_fops)) {
    printk("logi32.c: cannot register major number\n");
    return -EIO;
  }
  if (check_4_scanner())
  { printk("logi32.c: scanner interface not found\n");
    return -EIO;
  }
  return 0;
}

void cleanup_module(void)
{
  printk("Removing LOGITECH handheld scanner driver from memory\n");
  if(MOD_IN_USE)
    printk("logi32.c: device is busy, delaying driver removal\n");
  if(unregister_chrdev(LOGI32_SCANNER_MAJOR, LOGI32_SCANNER_NAME))
    printk("logi32.c: unregister_chrdev() failed\n");
}
